/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#include "simple.h"
#include "token.h"

void
shipout(token_list *tl,token tk)
     /* Append the token tk to the token list tl.
      * If tl==NULL then output the token. */
{
  if (!tl) outputtoken(tk);
  else appendtoken(tl,tk);
}

static token_list input_list={NULL,0,0};
/* This contains tokens that were generated by macro
 * expansion and are to be reprocessed. */

token
shipin(void)
     /* Read a token from the input list or the
      * input file if the input list is empty. */
{
  token tk;
  if (isempty(&input_list))
    return inputtoken();
  else {
    tk=firsttoken(&input_list);
    removefirst(&input_list);
    return tk;
  }
}

void expandtoken(token_list *tl,token tk);

void
expandargs(int *argc,token_list **argv)
     /* Read in the arguments to a macro and expand them.
      * Return in argc the number of arguments and in argv
      * a malloc()ated array of token lists corresponding
      * to the various arguments (after expansion of course). */
{
  token tk;
  *argc=1;
  *argv=malloc(sizeof(token_list));
  while (1) {
    makeemptylist(*argv+(*argc-1));
    while (1) {
      tk=shipin();
      if (tk==EOF)
	fatal(0,"Unterminated argument list.");
      if ((tk==END_ARGLIST)||(tk==NEXT_ARG))
	break;
      expandtoken(*argv+(*argc-1),tk);
    }
    if (tk==END_ARGLIST)
      break;
    (*argc)++;
    if ((*argv=realloc(*argv,*argc*sizeof(token_list)))==NULL)
      fatal(0,"Out of memory!");
  }
}

void
callmacro(token_list *expansion,token tk,int argc,token_list *argv)
     /* Call a (user or builtin) macro given by token tk,
      * with argc arguments given by argv, and put its
      * expansion in expansion. */
{
  if (argc<1)
    fatal(1,"argc<1");
  if (tk<FIRST_USER_COMMAND)
    callbuiltin(expansion,tk,argc,argv);
  else
    calluser(expansion,tk,argc,argv);
}

void
expandmacro(token tk)
     /* Read the arguments to a macro and expand it, putting
      * the expansion back on the input list. */
{
  int argc;
  token_list *argv;
  token_list expansion;
  expandargs(&argc,&argv);
  makeemptylist(&expansion);
  callmacro(&expansion,tk,argc,argv);
  for (;argc>0;argc--) freelist(argv+argc-1);
  free(argv);
  prependlist(&input_list,&expansion);
  freelist(&expansion);
}

void
expandtoken(token_list *tl,token tk)
     /* Expand the token tk to the token list tl.
      * If tl==NULL expand to the output. */
{
  int quotelevel;
  if ((tk==EOF)||(tk<FIRST_SPECIAL))
    shipout(tl,tk);
  else if (tk==QUOTE_NEXT) {
    tk=shipin();
    shipout(tl,tk);
  } else if (tk==OPEN_QUOTE) {
    quotelevel=1;
    while (quotelevel>=1) {
      tk=shipin();
      if (tk==OPEN_QUOTE)
	quotelevel++;
      else if (tk==CLOSE_QUOTE)
	quotelevel--;
      else if (tk==QUOTE_NEXT)
	tk=shipin();
      if (quotelevel>=1)
	shipout(tl,tk);
    }
  } else if (tk<FIRST_COMMAND)
    fatal(0,"Misplaced special token.");
  else
    expandmacro(tk);
  /* Note that macro expansion goes to the INPUT LIST
   * not to the tl token list, so it will be reread. */
}

void
expandinput(void)
     /* This is the main program loop. */
     /* Read the input and expand tokens as we go. */
{
  while (1)
    expandtoken(NULL,shipin());
}
