/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file is the global header file, containing definitions that
 * are needed about everywhere and prototypes for public functions.
 */

/* Version identifiers */

#define PRODUCTSTRING "SIMPLE"
#define VERSIONSTRING "1.1.0"
#define DATESTRING "17JAN98"

/* Essential typedefs */

typedef int token;
/* We define the token to be an ordinal type because they are _much_
 * easier to handle. This has some drawbacks, however; for example,
 * the number of user-defined commands is limited. */

typedef struct token_list {
  token *l; /* Pointer to the actual list of tokens. */
  /* This pointer is always malloc()ated, and realloc()ated when
   * the need arises. */
  int size; /* Size of the list (number of tokens in it). */
  int allsize; /* Allocated size (also in tokens, _not_ bytes). */
} token_list;

/* Token values. */
/* The actual values are unimportant. However, the order of the
 * sections is important (because to test whether a token belongs
 * to such or such a type we use tests such as `if (t>=FIRST_SPECIAL)' */

/* Note that token value -1 is reserved for EOF. No other negative
 * value is used. Also, value zero in each section (that is, 256,
 * 2048, 4096 and so on) is always reserved (who knows, it may turn
 * out to be useful some day). */

#define FIRST_SPECIAL 256
/* If we had a version which works with unicode we would change that
 * to 65536, I guess. */

/* These are the very special tokens. */
#define END_ARGLIST (FIRST_SPECIAL+1) /* " */
#define NEXT_ARG (FIRST_SPECIAL+2)    /* | */
#define OPEN_QUOTE (FIRST_SPECIAL+3)  /* < */
#define CLOSE_QUOTE (FIRST_SPECIAL+4) /* > */
#define QUOTE_NEXT (FIRST_SPECIAL+5)  /* # */

#define FIRST_PARAM 2048
/* Number of parameters is thus effectively limited to 2047 - sorry */
#define MAX_PARAM (FIRST_COMMAND-FIRST_PARAM-1)

#define FIRST_COMMAND 4096

/* These are the token numbers of builtin functions */
#define BI_IDENTITY (FIRST_COMMAND+1)
#define BI_VOID (FIRST_COMMAND+2)
#define BI_OUTPUT (FIRST_COMMAND+3)
#define BI_DEFINE (FIRST_COMMAND+4)
#define BI_INCLUDE (FIRST_COMMAND+5)
#define BI_TOKENIZE (FIRST_COMMAND+6)
#define BI_QUOTE (FIRST_COMMAND+7)
#define BI_IF (FIRST_COMMAND+8)
#define BI_HEAD (FIRST_COMMAND+9)
#define BI_TAIL (FIRST_COMMAND+10)
#define BI_AHEAD (FIRST_COMMAND+11)
#define BI_ATAIL (FIRST_COMMAND+12)

#define FIRST_USER_COMMAND 8192
/* Number of user-defined commands is thus effectively limited to 2 billion */

void fatal(char internal,char *fmt, ...);

void pushinputfile(char *name);

token identifycommand(char *name);

token inputtoken(void);

void outputtoken(token tk);

void expandinput(void);

void callbuiltin(token_list *expansion,token tk,int argc,token_list *argv);

void calluser(token_list *expansion,token tk,int argc,token_list *argv);

void defineuser(token_list *definition,token tk);
