/* SIMPLE - Simple Is a Macro Processing Language Element */
/* Copyright (c) 1998 David A. Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Handle user-defined macros (defining them, calling them, etc.).
 */

#include <stdio.h>
#include <stdlib.h>

#include "simple.h"
#include "token.h"

typedef struct def_entry {
  token who; /* What token is being defined */
  token_list what; /* Definition of that token */
} def_entry;

static def_entry *user_table = NULL;
static int user_table_len = 0;

void
replaceuser(int entry,token_list *definition)
     /* Modify the definition of a user token */
{
  freelist(&user_table[entry].what);
  copylist(&user_table[entry].what,definition);
}

void
createuser(token_list *definition,token tk)
     /* Create a new user token */
{
  if ((user_table=realloc(user_table,(++user_table_len)*sizeof(def_entry)))
      ==NULL)
    fatal(0,"Out of memory!");
  user_table[user_table_len-1].who=tk;
  makeemptylist(&user_table[user_table_len-1].what);
  replaceuser(user_table_len-1,definition);
}

void
defineuser(token_list *definition,token tk)
     /* Define a user token, checking to see whether it already exists */
{
  int i;
  for (i=0;i<user_table_len;i++) {
    if (tk==user_table[i].who) {
      replaceuser(i,definition);
      return;
    }
  }
  createuser(definition,tk);
}

void
expanduser(token_list *expansion,token_list *definition,
	   token tk,int argc,token_list *argv)
     /* Expand user token tk into definition, putting the result in
      * expansion, the number of arguments being argc and their
      * values given by argv. */
{
  int i,j;
  for (i=0;i<definition->size;i++) {
    if ((definition->l[i]>=FIRST_PARAM)&&(definition->l[i]<FIRST_COMMAND)) {
      j=definition->l[i]-FIRST_PARAM;
      if (j>argc) /* Append the empty list : no-op. */;
      else if (j==0) appendtoken(expansion,tk);
      else appendlist(expansion,argv+(j-1));
    } else appendtoken(expansion,definition->l[i]);
  }
}

void
calluser(token_list *expansion,token tk,int argc,token_list *argv)
     /* Expand user token tk looking up its definition, and put the
      * result in expansion, the number of arguments being argc and
      * their values given by argv. */
{
  int i;
  for (i=0;i<user_table_len;i++) {
    if (tk==user_table[i].who) {
      expanduser(expansion,&user_table[i].what,tk,argc,argv);
      return;
    }
  }
  fatal(0,"Undefined command.");
}
